/***********************************************************************\
**
**                          Copyright (c) 2008-2022 by
**                          HENGSTLER GmbH, Aldingen
**
**      This  software  is furnished under a license and may be used and
**      copied only in accordance with the terms  of  such  license  and
**      with the inclusion of the above copyright notice.  This software
**      or any other copies thereof may not  be  provided  or  otherwise
**      made  available  to any other person.  No title to and ownership
**      of the software is hereby transferred.
**
**      The  information  in  this software is subject to change without
**      notice and should not be construed as a commitment by Hengstler.
**
**  Module:      Xpm_Pmp_Interface.cpp
**  Facility:    S690001 - Hengstler XPM Printer Driver for Windows XP/7/8/10
**  Component:   Print Monitor Plugin (Xpm-PrintMonPlugin)
**  Version:     R8-V1.07-FINAL / 2022OCT18
**  Environment: Windows XP/7/8/10
**  Compiler:    Microsoft Visual Studio 2015
**  Abstract:    Plugin interface to print monitor 
**  Note:        -
**  History:     R1-V1.00c / 2008JUL11 / HOR / 1st RELEASE - 1st BETA
**               R1-V1.00d / 2008DEC11 / HOR / 1st RELEASE - 2nd BETA
**               R1-V1.00e / 2009MAY05 / HOR / 1st RELEASE - 3rd BETA
**               R1-V1.00f / 2009JUL16 / HOR / 1st RELEASE - 4th BETA
**               R1-V1.00  / 2009AUG27 / HOR / 1st RELEASE - FINAL 
**               ------------------------------------------------------------
**               R2-V1.01a / 2010JAN26 / HOR / 2nd RELEASE - 1st BETA 
**               R2-V1.01b / 2010MAR04 / HOR / 2nd RELEASE - 2nd BETA
**               R2-V1.01c / 2011DEC19 / HOR / M-R2-012-HOR
**                  Complete driver package converted to VS2010 and adapted to Windows 7 32/64 bit
**               R2-V1.01c / 2011DEC19 / HOR / 2nd RELEASE - 3rd BETA
**               R2-V1.01d / 2012JUN27 / HOR / 2nd RELEASE - 4th BETA
**               R2-V1.01e / 2015MAY05 / HOR / 2nd RELEASE - 5th BETA
**               R2-V1.01f / 2016FEB15 / HOR / 2nd RELEASE - 6th BETA
**               R2-V1.01  / 2016AUG19 / HOR / M-R2-041-HOR
**                 Complete driver package converted to VS2015 and adapted to Windows 8/10 32/64 bit
**               R2-V1.01  / 2016AUG19 / HOR / 2nd RELEASE - FINAL
**               ------------------------------------------------------------
**               R3-V1.02  / 2016AUG24 / HOR / M-R3-001-HOR
**                 Additional driver package compilation added for usage within Windows-XP.
**               R3-V1.02  / 2016AUG24 / HOR / 3rd RELEASE - FINAL
**               ------------------------------------------------------------
**               R4-V1.03  / 2018JUL10 / HOR / 4th RELEASE - FINAL
**               ------------------------------------------------------------
**               R5-V1.04  / 2018OCT11 / HOR / 5th RELEASE - FINAL
**               ------------------------------------------------------------
**               R6-V1.05  / 2018DEC07 / HOR / 6th RELEASE - FINAL
**               ------------------------------------------------------------
**               R7-V1.06  / 2022FEB10 / HOR / 7th RELEASE - FINAL
**               ------------------------------------------------------------
**               R8-V1.07  / 2022OCT18 / HOR / 8th RELEASE - FINAL
**               ------------------------------------------------------------
**
\**********************************************************************/

/*
** INCLUDES
*/

#include "Xpm_Pmp_System.h"
#include <ctype.h>
#include "Xpm_Pmp_Main.h"
#include "Xpm_Pmp_JobInfo.h"
#include "Xpm_Pmp_PrinterInfo.h"
#include "Xpm_Pmp_PrinterStatus.h"
#include "Xpm_Pmp_PrinterSensor.h"
#include "Xpm_Pmp_PrinterPresent.h"
#include "Xpm_Pmp_PrinterFeed.h"
#include "Xpm_PrintMonPlugin.h"
#include "Xpm_Pmp_RegistryAccess.h"
#include "Xpm_Pmp_LogMan.h"
#include "Xpm_Pmp_Interface.h"

/*
** COMPILER
*/

#ifdef _DEBUG
// #define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/*
** MACROS 
*/

//-- Macros specifying the default delays for printer data inquiry (e.g. status requests) 
#define D_XPM_PMP_PRINTER_INQUIRY_DELAY_START_MS     50  /* Delay at start of data printer inquiry, before writing request to printer */
#define D_XPM_PMP_PRINTER_INQUIRY_DELAY_RESPONSE_MS 150  /* Delay during of printer data inquiry, after writing request to printer */
#define D_XPM_PMP_PRINTER_INQUIRY_DELAY_END_MS       50  /* Delay at end of printer data inquiry, after having received printer data */

/*
** TYPES
*/

/*
** FUNCTIONS
*/

bool Xpm_Pmp_Ifc_GetPluginSettings   (P_XPM_PLUGIN_DATA psData);
bool Xpm_Pmp_Ifc_SynchronizePrinter  (P_XPM_PLUGIN_DATA psData);

/*
** VARIABLES
*/

//********************************************************************************************************
//** FUNCTION     Xpm_Pmp_StartMonitor 
//** ABSTRACT     Printer monitor plugin function for start of monitoring
//**              This function is called by the printer monitor after the port has been opened 
//** PARAMETER    pszPortName           Name of the port
//**              pszPrinterName        Name of the printer
//**              hPmCallback           Handle for port monitor callbacks
//**              pdwServiceIntervalSec Required service interval in sec (min 30, disabled 0)
//**              pfnWritePort          Callback function for writing data to port
//**              pfnReadPort           Callback function for reading data from port
//**              pfnCancelDocument     Callback function for cancelling the document
//** RETURN       hData                 Allocated data that is used by other ports (will be freed in ...ClosePort) 
//** NOTE         Type must be equal to TF_Xpm_Pmp_StartMonitor in "Xpm_PrintMonPlugin.h"
//********************************************************************************************************

__declspec(dllexport) HANDLE Xpm_Pmp_StartMonitor
   ( LPTSTR pszPortName, LPTSTR pszPrinterName, HANDLE hPmCallback, DWORD * pdwServiceIntervalSec
   , BOOL (*pfnWritePort)  (HANDLE, LPBYTE, DWORD, LPDWORD, DWORD)
   , BOOL (*pfnReadPort)   (HANDLE, LPBYTE, DWORD, LPDWORD, DWORD)
   , VOID (*pfnCancelDocument)  (HANDLE)
   )
{
   P_XPM_PLUGIN_DATA psData  = NULL;  // Data allocated for port I/O
   HANDLE            hResult = NULL;  // Return value, no data by default

   //-- The debug break will be of need when the print monitor plugin 
   //.. is to be debugged with use of visual studio
   // DebugBreak();

   //-- Function type test
   TF_Xpm_Pmp_StartMonitor pFunctionTypeTest = Xpm_Pmp_StartMonitor;

   //-- Check parameter
   if (!pszPortName || !pszPrinterName)
      goto FAILURE;

   //-- Allocate data plugin data
   if ((psData = (P_XPM_PLUGIN_DATA) malloc(sizeof(S_XPM_PLUGIN_DATA))) == NULL)
      goto FAILURE;

   //-- Initialize structure
   psData->dwSructId         = P_XPM_PLUGIN_DATA_ID;
   psData->pszPortName       = NULL; 
   psData->pszPrinterName    = NULL; 
   psData->pszPrinterNameDoc = NULL; 
   psData->hPmCallback       = NULL;
   psData->pfnWritePort      = NULL;
   psData->pfnReadPort       = NULL;
   psData->dwJobId           = (DWORD) -1;

   //-- Set read write callback function pointers and port handle
   psData->hPmCallback       = hPmCallback;
   psData->pfnWritePort      = pfnWritePort;
   psData->pfnReadPort       = pfnReadPort;
   psData->pfnCancelDocument = pfnCancelDocument;

   //-- Save port and printer name
   if (pszPortName) {
      if ((psData->pszPortName = (LPTSTR) malloc(wcslen(pszPortName)*2+2)) != NULL)
         wcscpy(psData->pszPortName, pszPortName);
   } 
   if (pszPrinterName) {
      if ((psData->pszPrinterName = (LPTSTR) malloc(wcslen(pszPrinterName)*2+2)) != NULL)
         wcscpy(psData->pszPrinterName, pszPrinterName);
   } 

   //-- Get plugin settings 
   Xpm_Pmp_Ifc_GetPluginSettings(psData);

   //-- Set the service interval in seconds for periodical calls to Xpm_Pmp_ServiceRoutine(..)
   //.. Note that minimum is 5 seconds and that 0 sec disables service calls
   if (pdwServiceIntervalSec) {
      *pdwServiceIntervalSec = psData->sSettings.wServiceIntervalSec;
      if ((*pdwServiceIntervalSec != 0) && (*pdwServiceIntervalSec < 5))
         *pdwServiceIntervalSec = 5;
   }

   //-- Correct maximum duration of a page printout [30sec..30min]
   psData->sSettings.wMaxDurationOfPrintSec = min(max(psData->sSettings.wMaxDurationOfPrintSec, 30), 1800);

   //-- Initialize printer info in registry
   Xpm_Pmp_Pi_InitializePrinterInfo(psData);

   //-- Initialize job info in registry
   Xpm_Pmp_Ji_InitializeJobInfo(psData);

   //-- Initialize printer status in registry
   Xpm_Pmp_Pst_InitializePrinterStatus(psData);

   //-- Initialize printer sensor in registry
   Xpm_Pmp_Pse_InitializePrinterSensor(psData);

   //-- Initialize printer present in registry
   Xpm_Pmp_Ppr_InitializePrinterPresent(psData);

   //-- Initialize printer feed in registry
   Xpm_Pmp_Pfe_InitializePrinterFeed(psData);

   //-- Create log manager
   psData->poXpmPmpLogMan = new CXpm_Pmp_LogMan;

   //-- Set return handle to pointer to private data 
   hResult = (HANDLE) psData;
     
   //-- All done
   goto EXIT;

FAILURE:

   //-- Free all allocated data
   if (psData) {
      if (psData->pszPortName)    free(psData->pszPortName);    psData->pszPortName    = NULL;
      if (psData->pszPrinterName) free(psData->pszPrinterName); psData->pszPrinterName = NULL;
      free(psData); 
      psData = NULL;
   }

EXIT:
 
   //-- Return handle to private data on success and NULL otherwise
   return(hResult);
}

//********************************************************************************************************
//** FUNCTION     Xpm_Pmp_EndMonitor 
//** ABSTRACT     Printer monitor plugin function for end of monitoring
//**              This function is called by the printer monitor after the port has been closed 
//** PARAMETER    hData         Handle pointing to private plugin data (from Xpm_Pmp_StartMonitor)
//** RETURN       fResult       Flag indicating success if TRUE and false otherwise 
//** NOTE         Type must be equal to TF_Xpm_Pmp_EndMonitor in "Xpm_PrintMonPlugin.h"
//********************************************************************************************************

__declspec(dllexport) BOOL Xpm_Pmp_EndMonitor                    
   (HANDLE hData)
{
   P_XPM_PLUGIN_DATA psData = (P_XPM_PLUGIN_DATA) hData;  // Data allocated for port I/O
   BOOL              fResult = FALSE;                        // Return value, error by default

   //-- Function type test
   TF_Xpm_Pmp_EndMonitor pFunctionTypeTest = Xpm_Pmp_EndMonitor;

   //-- Check parameter
   if (!psData || (psData->dwSructId != P_XPM_PLUGIN_DATA_ID))
      goto FAILURE;

   //-- Destroy log manager
   if (psData->poXpmPmpLogMan) delete psData->poXpmPmpLogMan;

   //-- Free allocated plugin data 
   if (psData->pszPortName)    free(psData->pszPortName);    psData->pszPortName    = NULL;
   if (psData->pszPrinterName) free(psData->pszPrinterName); psData->pszPrinterName = NULL;
   free(psData); psData = NULL;
 
   //-- Function has been succesfull
   fResult = TRUE;
     
   //-- All done
   goto EXIT;

FAILURE:

EXIT: 
 
   //-- Return TRUE on success and FALSE otherwise
   return(fResult);
}

//********************************************************************************************************
//** FUNCTION     Xpm_Pmp_StartDocument
//** ABSTRACT     Printer monitor plugin function for start of document (start of print job)
//**              This function is called by the printer monitor after the document printout has been started
//** PARAMETER    hData          Handle pointing to private plugin data (from Xpm_Pmp_StartMonitor)
//**              pszPrinterName Name of the printer
//**              dwJobId        Id of job that is being done
//**              dwDocInfoLevel Level of DOC_INFO_X structure (DOC_INFO_1 or DOC_INFO_2)
//**              pbDocInfo      Pointer to DOC_INFO_X structure (DOC_INFO_1 or DOC_INFO_2)
//** RETURN       fResult        Flag indicating success if TRUE and false otherwise 
//** NOTE         Type must be equal to TF_Xpm_Pmp_StartDocument in "Xpm_PrintMonPlugin.h"
//********************************************************************************************************

__declspec(dllexport) BOOL Xpm_Pmp_StartDocument
   (HANDLE hData, LPTSTR pszPrinterName, DWORD dwJobId, DWORD dwDocInfoLevel, LPBYTE  pbDocInfo)
{
   P_XPM_PLUGIN_DATA psData = (P_XPM_PLUGIN_DATA) hData;  // Data allocated for port I/O
   BOOL              fResult = FALSE;                        // Return value, error by default

   //DebugBreak();

   //-- Function type test
   TF_Xpm_Pmp_StartDocument pFunctionTypeTest = Xpm_Pmp_StartDocument;

   //-- Validate parameter
   if (!psData || (psData->dwSructId != P_XPM_PLUGIN_DATA_ID))
      goto FAILURE;

   //-- Store job-id
   psData->dwJobId = dwJobId;

   //-- Free printer name (if defined)
   if (psData->pszPrinterNameDoc) free(psData->pszPrinterNameDoc); psData->pszPrinterNameDoc = NULL;

   //-- Store printer name (if allocated)
   if (pszPrinterName) {
      if ((psData->pszPrinterNameDoc = (LPTSTR) malloc(wcslen(pszPrinterName)*2+2)) != NULL)
         wcscpy(psData->pszPrinterNameDoc, pszPrinterName);
   } 

   //-- Update job information in registry
   Xpm_Pmp_Ji_UpdateJobInfo_JobStarted(psData,  psData->dwJobId);

   //-- Function has been succesfull
   fResult = TRUE;
 
   //-- All done
   goto EXIT;

FAILURE:

EXIT:
 
   //-- Return TRUE on success and FALSE otherwise
   return(fResult);
}

//********************************************************************************************************
//** FUNCTION     Xpm_Pmp_EndDocument
//** ABSTRACT     Printer monitor plugin function for end of document (end of print job)
//**              This function is called by the printer monitor before the document printout has ended
//** PARAMETER    hData         Handle pointing to private plugin data (from Xpm_Pmp_StartMonitor)
//** RETURN       fResult       Flag indicating success if TRUE and false otherwise 
//** NOTE         Type must be equal to TF_Xpm_Pmp_EndDocument in "Xpm_PrintMonPlugin.h"
//********************************************************************************************************

__declspec(dllexport) BOOL Xpm_Pmp_EndDocument
   (HANDLE hData)
{
   P_XPM_PLUGIN_DATA psData              = (P_XPM_PLUGIN_DATA) hData;  // Data allocated for port I/O
   BOOL              fResult             = FALSE;                      // Return value, error by default
   bool              fPrinterMightBeBusy = true;                       // Flag indicating that printer is busy printing
   bool              fConnected          = true;                       // Flag indicating that printer is connected

   //-- Function type test
   TF_Xpm_Pmp_EndDocument pFunctionTypeTest = Xpm_Pmp_EndDocument;

   //-- Validate parameter
   if (!psData || (psData->dwSructId != P_XPM_PLUGIN_DATA_ID))
      goto FAILURE;

   //-- Synchronize printer (read pending data and send synchronization bytes)
   if (Xpm_Pmp_Ifc_SynchronizePrinter(psData)) {

      //-- Update printer information in registry
      if (!(psData->sSettings.dwOptions & D_XPM_PMP_SET_OPT_UPDATE_PRINTER_INFO_AT_END_OF_DOC_DISABLED)) {
          fConnected = Xpm_Pmp_Pi_UpdatePrinterInfo(psData, fPrinterMightBeBusy, !fConnected);
          fPrinterMightBeBusy = false;
      }
   
      //-- Update printer status in registry
      if (!(psData->sSettings.dwOptions & D_XPM_PMP_SET_OPT_UPDATE_PRINTER_STATUS_AT_END_OF_DOC_DISABLED)) {
         fConnected = Xpm_Pmp_Pst_UpdatePrinterStatus(psData, fPrinterMightBeBusy, !fConnected);
         fPrinterMightBeBusy = false;
      }

      //-- Update printer sensor in registry
      if (!(psData->sSettings.dwOptions & D_XPM_PMP_SET_OPT_UPDATE_PRINTER_SENSOR_AT_END_OF_DOC_DISABLED)) {
         fConnected = Xpm_Pmp_Pse_UpdatePrinterSensor(psData, fPrinterMightBeBusy, !fConnected);
         fPrinterMightBeBusy = false;
      }

      //-- Update printer present in registry
      if (!(psData->sSettings.dwOptions & D_XPM_PMP_SET_OPT_UPDATE_PRINTER_PRESENT_AT_END_OF_DOC_DISABLED)) {
         fConnected = Xpm_Pmp_Ppr_UpdatePrinterPresent(psData, fPrinterMightBeBusy, !fConnected);
         fPrinterMightBeBusy = false;
      }

      //-- Update printer feed in registry
      if (!(psData->sSettings.dwOptions & D_XPM_PMP_SET_OPT_UPDATE_PRINTER_FEED_AT_END_OF_DOC_DISABLED)) {
         fConnected = Xpm_Pmp_Pfe_UpdatePrinterFeed(psData, fPrinterMightBeBusy, !fConnected);
         fPrinterMightBeBusy = false;
      }
   }

   //-- Update job information in registry
   Xpm_Pmp_Ji_UpdateJobInfo_JobCompleted(psData,  psData->dwJobId);
 
   //-- Reset job-id
   psData->dwJobId = (DWORD) -1;

   //-- Free printer name (if allocated)
   if (psData->pszPrinterNameDoc) free(psData->pszPrinterNameDoc); psData->pszPrinterNameDoc = NULL;
    
   //-- Function has been succesfull
   fResult = TRUE;

   //-- All done
   goto EXIT;

FAILURE:

EXIT:
 
   //-- Return TRUE on success and FALSE otherwise
   return(fResult);
}

//********************************************************************************************************
//** FUNCTION     Xpm_Pmp_CancelDocument
//** ABSTRACT     Printer monitor plugin function for cancellation of documents 
//**              This function is called by the printer monitor whenever a write to the port is unsuccessful.
//** PARAMETER    hData         Handle pointing to private plugin data (from Xpm_Pmp_StartMonitor)
//** RETURN       fResult       Flag indicating cancel printing if TRUE and continue document otherwise 
//** NOTE         Type must be equal to TF_Xpm_Pmp_CancelDocument in "Xpm_PrintMonPlugin.h"
//********************************************************************************************************

__declspec(dllexport) BOOL Xpm_Pmp_CancelDocument
   (HANDLE hData, DWORD dwDurationOfPrintSec)
{
   P_XPM_PLUGIN_DATA psData = (P_XPM_PLUGIN_DATA) hData;  // Data allocated for port I/O
   BOOL              fResult = FALSE;                        // Return value, error by default

   //-- Function type test
   TF_Xpm_Pmp_CancelDocument pFunctionTypeTest = Xpm_Pmp_CancelDocument;

   //-- Validate parameter
   if (!psData || (psData->dwSructId != P_XPM_PLUGIN_DATA_ID))
      goto FAILURE;

   //-- Check if age exceeds maximum 
   if (dwDurationOfPrintSec < psData->sSettings.wMaxDurationOfPrintSec) 
      goto EXIT;

   //-- Update job information in registry 
   Xpm_Pmp_Ji_UpdateJobInfo_JobCancelled(psData, psData->dwJobId);
 
   //-- Reset job-id
   psData->dwJobId = (DWORD) -1;

   //-- Free printer name (if allocated)
   if (psData->pszPrinterNameDoc) free(psData->pszPrinterNameDoc); psData->pszPrinterNameDoc = NULL;
    
   //-- Function has been succesfull 
   //.. Return value will cause the print job to be cancelled
   fResult = TRUE;

   //-- All done
   goto EXIT;

FAILURE:

EXIT:
 
   //-- Return TRUE on success and FALSE otherwise
   return(fResult);
}

//********************************************************************************************************
//** FUNCTION     Xpm_Pmp_StartPage
//** ABSTRACT     Printer monitor plugin function for start of page (in a document)
//**              This function is called by the printer monitor at start of the page printout 
//** PARAMETER    hData          Handle pointing to private plugin data (from Xpm_Pmp_StartMonitor)
//** RETURN       fResult        Flag indicating success if TRUE and false otherwise 
//** NOTE         Type must be equal to TF_Xpm_Pmp_StartDocument in "Xpm_PrintMonPlugin.h"
//********************************************************************************************************

__declspec(dllexport) BOOL Xpm_Pmp_StartPage
   (HANDLE hData)
{
   P_XPM_PLUGIN_DATA psData = (P_XPM_PLUGIN_DATA) hData;  // Data allocated for port I/O
   BOOL              fResult = FALSE;                        // Return value, error by default

   //-- Function type test
   TF_Xpm_Pmp_StartPage pFunctionTypeTest = Xpm_Pmp_StartPage;

   //-- Validate parameter
   if (!psData || (psData->dwSructId != P_XPM_PLUGIN_DATA_ID))
      goto FAILURE;

   //-- Function has been succesfull
   fResult = TRUE;
 
   //-- All done
   goto EXIT;

FAILURE:

EXIT:
 
   //-- Return TRUE on success and FALSE otherwise
   return(fResult);
}

//********************************************************************************************************
//** FUNCTION     Xpm_Pmp_EndPage
//** ABSTRACT     Printer monitor plugin function for end of page (in a document)
//**              This function is called by the printer monitor after a page has been printed 
//** PARAMETER    hData         Handle pointing to private plugin data (from Xpm_Pmp_StartMonitor)
//** RETURN       fResult       Flag indicating success if TRUE and false otherwise 
//** NOTE         Type must be equal to TF_Xpm_Pmp_EndDocument in "Xpm_PrintMonPlugin.h"
//********************************************************************************************************

__declspec(dllexport) BOOL Xpm_Pmp_EndPage
   (HANDLE hData)
{
   P_XPM_PLUGIN_DATA psData              = (P_XPM_PLUGIN_DATA) hData;  // Data allocated for port I/O
   BOOL              fResult             = FALSE;                      // Return value, error by default
   bool              fPrinterMightBeBusy = true;                       // Flag indicating that printer is busy printing
   bool              fConnected          = true;                       // Flag indicating that printer is connected

   //-- Function type test
   TF_Xpm_Pmp_EndPage pFunctionTypeTest = Xpm_Pmp_EndPage;

   //-- Validate parameter
   if (!psData || (psData->dwSructId != P_XPM_PLUGIN_DATA_ID))
      goto FAILURE;

   //-- Synchronize printer (read pending data and send synchronization bytes)
   if (Xpm_Pmp_Ifc_SynchronizePrinter(psData)) {

      //-- Update printer information in registry
      if (!(psData->sSettings.dwOptions & D_XPM_PMP_SET_OPT_UPDATE_PRINTER_INFO_AT_END_OF_PAGE_DISABLED)) {
          fConnected = Xpm_Pmp_Pi_UpdatePrinterInfo(psData, fPrinterMightBeBusy, !fConnected);
          fPrinterMightBeBusy = false;
      }
   
      //-- Update printer status in registry
      if (!(psData->sSettings.dwOptions & D_XPM_PMP_SET_OPT_UPDATE_PRINTER_STATUS_AT_END_OF_PAGE_DISABLED)) {
         fConnected = Xpm_Pmp_Pst_UpdatePrinterStatus(psData, fPrinterMightBeBusy, !fConnected);
         fPrinterMightBeBusy = false;
      }

      //-- Update printer sensor in registry
      if (!(psData->sSettings.dwOptions & D_XPM_PMP_SET_OPT_UPDATE_PRINTER_SENSOR_AT_END_OF_PAGE_DISABLED)) {
         fConnected = Xpm_Pmp_Pse_UpdatePrinterSensor(psData, fPrinterMightBeBusy, !fConnected);
         fPrinterMightBeBusy = false;
      }

      //-- Update printer present in registry
      if (!(psData->sSettings.dwOptions & D_XPM_PMP_SET_OPT_UPDATE_PRINTER_PRESENT_AT_END_OF_PAGE_DISABLED)) {
         fConnected = Xpm_Pmp_Ppr_UpdatePrinterPresent(psData, fPrinterMightBeBusy, !fConnected);
         fPrinterMightBeBusy = false;
      }

      //-- Update printer feed in registry
      if (!(psData->sSettings.dwOptions & D_XPM_PMP_SET_OPT_UPDATE_PRINTER_FEED_AT_END_OF_PAGE_DISABLED)) {
         fConnected = Xpm_Pmp_Pfe_UpdatePrinterFeed(psData, fPrinterMightBeBusy, !fConnected);
         fPrinterMightBeBusy = false;
      }
   }

   //-- Check if presenter monitoring is not disabled
   if (!(psData->sSettings.dwOptions & D_XPM_PMP_SET_OPT_MONITOR_PRESENTER_AT_END_OF_PAGE_DISABLED)) {

      //-- Get timestamp at start of monitoring
      DWORD dwTimeStampMs = GetTickCount();

      //-- Continue until presenter is ready
      while (true) {

         TS_XPM_PMP_PRINTER_PRESENT  sPrinterPresent; // Printer present packet

         //-- Update printer present in registry
         if ((fConnected = Xpm_Pmp_Ppr_UpdatePrinterPresent(psData, fPrinterMightBeBusy, !fConnected)) == false) 
            break;

         //-- Read printer present status from registry
         if (!Xpm_Pmp_Ra_ReadPrinterDataFromRegistry(psData, D_XPM_PMP_REG_VALUE_PMP_PRINTER_PRESENT, (PBYTE) &sPrinterPresent, sizeof(sPrinterPresent)))
            goto FAILURE;

         //-- Check if a loop/retract presenter is connected
         if (sPrinterPresent.sPrinterPresent.ubType != 3) 
            break;

         //-- Check if the presenter is:
         //.. not busy presenting printouts
         //.. not busy rejecting printouts
         //.. and has rejecting printouts
         if (    !(sPrinterPresent.sPrinterPresent.wStatus & D_XPM_API_PPSW_PRESENTING_PRINTOUTS)
              && !(sPrinterPresent.sPrinterPresent.wStatus & D_XPM_API_PPSW_REJECTING_PRINTOUTS) 
              && (sPrinterPresent.sPrinterPresent.ubNrOfLoadedPrintouts == 0) 
            ) {
       
            //-- Cancel document if printouts are rejected
            if (sPrinterPresent.sPrinterPresent.wStatus & D_XPM_API_PPSW_PRINTOUTS_REJECTED) 
               psData->pfnCancelDocument(psData->hPmCallback);

            //-- Presenting completed
            break;
         }     

         //-- Determine current time 
         DWORD dwTimeMs = GetTickCount();

         //-- Determine duration of monitoring
         DWORD dwDurationMs = (dwTimeMs >= dwTimeStampMs ? dwTimeMs - dwTimeStampMs : (0xFFFFFFFF - dwTimeStampMs) + dwTimeMs);

         //-- Quit monitoring after ... seconds
         if (dwDurationMs > 60000) {

            //-- Cancel document
            psData->pfnCancelDocument(psData->hPmCallback);
 
            //-- Presenting completed
            break;
         }

         //-- Wait
         Sleep(250);
      }
   }

   //-- Update job information in registry
   Xpm_Pmp_Ji_UpdateJobInfo_PageCompleted(psData,  psData->dwJobId);

   //-- Function has been succesfull
   fResult = TRUE;

   //-- All done
   goto EXIT;

FAILURE:

EXIT:
 
   //-- Return TRUE on success and FALSE otherwise
   return(fResult);
}

//********************************************************************************************************
//** FUNCTION     Xpm_Pmp_ServiceRoutine
//** ABSTRACT     Printer monitor plugin function for periodical service call
//**              This function is called by the printer monitor every ... sec 
//** PARAMETER    hData                 Handle pointing to private plugin data (from Xpm_Pmp_StartMonitor)
//** RETURN       fResult               Flag indicating success if TRUE and false otherwise 
//** NOTE         Type must be equal to TF_Xpm_Pmp_ServiceRoutine in "Xpm_PrintMonPlugin.h"
//********************************************************************************************************

__declspec(dllexport) BOOL Xpm_Pmp_ServiceRoutine
   (HANDLE hData)
{
   P_XPM_PLUGIN_DATA psData     = (P_XPM_PLUGIN_DATA) hData;  // Data allocated for port I/O
   BOOL              fResult    = FALSE;                      // Return value, error by default
   bool              fConnected = true;                       // Flag indicating that printer is connected

   //-- Function type test
   TF_Xpm_Pmp_ServiceRoutine pFunctionTypeTest = Xpm_Pmp_ServiceRoutine;

   //-- Check parameter
   if (!psData || (psData->dwSructId != P_XPM_PLUGIN_DATA_ID))
      goto FAILURE;

   //-- Synchronize printer (read pending data and send synchronization bytes)
   if (!Xpm_Pmp_Ifc_SynchronizePrinter(psData)) 
      goto FAILURE;

   //-- Update printer information in registry
   if (!(psData->sSettings.dwOptions & D_XPM_PMP_SET_OPT_UPDATE_PRINTER_INFO_DURING_SERVICE_DISABLED))
      fConnected = Xpm_Pmp_Pi_UpdatePrinterInfo(psData, false, !fConnected);

   //-- Update printer status in registry
   if (!(psData->sSettings.dwOptions & D_XPM_PMP_SET_OPT_UPDATE_PRINTER_STATUS_DURING_SERVICE_DISABLED))
      fConnected = Xpm_Pmp_Pst_UpdatePrinterStatus(psData, false, !fConnected);
  
   //-- Update printer sensor in registry
   if (!(psData->sSettings.dwOptions & D_XPM_PMP_SET_OPT_UPDATE_PRINTER_SENSOR_DURING_SERVICE_DISABLED))
      fConnected = Xpm_Pmp_Pse_UpdatePrinterSensor(psData, false, !fConnected);

   //-- Update printer present in registry
   if (!(psData->sSettings.dwOptions & D_XPM_PMP_SET_OPT_UPDATE_PRINTER_PRESENT_DURING_SERVICE_DISABLED))
      fConnected = Xpm_Pmp_Ppr_UpdatePrinterPresent(psData, false, !fConnected);

   //-- Update printer feed in registry
   if (!(psData->sSettings.dwOptions & D_XPM_PMP_SET_OPT_UPDATE_PRINTER_FEED_DURING_SERVICE_DISABLED))
      fConnected = Xpm_Pmp_Pfe_UpdatePrinterFeed(psData, false, !fConnected);

   //-- Function has been succesfull
   fResult = TRUE;
     
   //-- All done
   goto EXIT;

FAILURE:

EXIT: 
 
   //-- Return TRUE on success and FALSE otherwise

   return(fResult);
}

//******************************************************************************************
// FUNCTION   Xpm_Pmp_Ifc_InquirePrinterData
// ABSTRACT   Inquires printer data
//******************************************************************************************

bool Xpm_Pmp_Ifc_InquirePrinterData
   (P_XPM_PLUGIN_DATA psData, BYTE ubId 
   ,LPBYTE pubRequest, DWORD dwRequest, LPBYTE pubResponse, DWORD dwResponse
   ,DWORD dwReadTimeoutMs)
{
   bool                     fResult             = false;  // Return value, FALSE by default
   DWORD                    dwWritten, dwRead;            // Nr of bytes written/read
   BYTE                     ubData;                       // Data byte
   TS_XPM_API_PCKT_HEADER * psPcktHeader  = (TS_XPM_API_PCKT_HEADER *) pubResponse; // Packet header pointer

   //-- Validate parameter
   if (!psData || (psData->dwSructId != P_XPM_PLUGIN_DATA_ID) || !pubRequest || !dwRequest || !pubResponse || ! dwResponse)
      goto FAILURE;

   //-- Read all buffered data 
   while (psData->pfnReadPort(psData->hPmCallback, &ubData, sizeof(ubData), &dwRead, 0) && (dwRead == sizeof(ubData)));

   //-- Sleep .. ms to allows the printer to recover from previous I/O before handling this inquiry
   Sleep(psData->sSettings.wDelayInquiryStartMs);

   //-- Write printer  request to printer 
   if (!psData->pfnWritePort(psData->hPmCallback, pubRequest, dwRequest, &dwWritten, 1000) || (dwWritten != dwRequest))
      goto FAILURE;
 
   //-- Sleep to give the printer time to answer 
   Sleep(psData->sSettings.wDelayInquiryResponseMs);

   //-- Receive printer status from printer
   if (!psData->pfnReadPort(psData->hPmCallback, pubResponse, dwResponse, &dwRead, dwReadTimeoutMs) || (dwRead != dwResponse))
      goto FAILURE;

   //-- Check for packet header being correct
   if (    (psPcktHeader->ubSync1                          != 0x1B)
        || (psPcktHeader->ubSync2                          != 0xFF)
        || (psPcktHeader->ubId                             != ubId)
        || (psPcktHeader->ubLength + sizeof(*psPcktHeader) != dwResponse)
      ) goto FAILURE;

   //-- Sleep .. ms to allows the printer to recover from this inquiry before handling other I/O 
   Sleep(psData->sSettings.wDelayInquiryEndMs);

   //-- Success
   fResult = true;

   //-- All done
   goto EXIT;

FAILURE:
   ;

EXIT:

   //-- Return TRUE on success and FALSE otherwise
   return(fResult);
}

//******************************************************************************************
// FUNCTION   Xpm_Pmp_Ifc_GetPluginSettings
// ABSTRACT   Reads plugin settings from registry
//******************************************************************************************

bool Xpm_Pmp_Ifc_GetPluginSettings
   (P_XPM_PLUGIN_DATA psData)
{
   bool                fResult    = false;  // Return value, FALSE by default
   
   //-- Check parameter
   if (!psData || (psData->dwSructId != P_XPM_PLUGIN_DATA_ID))
      goto FAILURE;

   //-- Read settings from registry
   if (    !Xpm_Pmp_Ra_ReadPrinterDataFromRegistry(psData, D_XPM_PMP_REG_VALUE_PMP_SETTINGS, (PBYTE) &psData->sSettings, sizeof(psData->sSettings))
        || (psData->sSettings.ubRevision != D_XPM_PMP_SET_REVISION)
      ) {

      //-- Initialize settings structure 
      memset(&psData->sSettings, 0, sizeof(psData->sSettings));           // Clear complete structure
      psData->sSettings.dwSize               = sizeof(psData->sSettings); // Set size of this structure
      psData->sSettings.ubRevision           = D_XPM_PMP_SET_REVISION;    // Set revision number (D_XPM_PMP_SET_REVISION)
      psData->sSettings.wServiceIntervalSec  = 30;                        // Set service intervall in seconds
   }

   //-- Correct contents of settings structure 
   if ((psData->sSettings.wServiceIntervalSec > 0) && (psData->sSettings.wServiceIntervalSec < 5))
       psData->sSettings.wServiceIntervalSec = 5;

   //-- Initialize and correct delay at start of data inquiry
   if (!psData->sSettings.wDelayInquiryStartMs)
      psData->sSettings.wDelayInquiryStartMs = D_XPM_PMP_PRINTER_INQUIRY_DELAY_START_MS;
   psData->sSettings.wDelayInquiryStartMs = min(max(psData->sSettings.wDelayInquiryStartMs, 1), 1000);

   //-- Initialize and correct delay before reading response during data inquiry
   if (!psData->sSettings.wDelayInquiryResponseMs)
      psData->sSettings.wDelayInquiryResponseMs = D_XPM_PMP_PRINTER_INQUIRY_DELAY_RESPONSE_MS;
   psData->sSettings.wDelayInquiryResponseMs = min(max(psData->sSettings.wDelayInquiryResponseMs, 1), 1000);

   //-- Initialize and correct delay at end of data inquiry
   if (!psData->sSettings.wDelayInquiryEndMs)
      psData->sSettings.wDelayInquiryEndMs = D_XPM_PMP_PRINTER_INQUIRY_DELAY_START_MS;
   psData->sSettings.wDelayInquiryEndMs = min(max(psData->sSettings.wDelayInquiryStartMs, 1), 1000);

   //-- Write settings to registry
   if (!Xpm_Pmp_Ra_WritePrinterDataToRegistry(psData, D_XPM_PMP_REG_VALUE_PMP_SETTINGS, (PBYTE) &psData->sSettings, sizeof(psData->sSettings)))
      goto FAILURE;

   //-- Success
   fResult = true;

   //-- All done
   goto EXIT;

FAILURE:
   ;

EXIT:

   //-- Return TRUE on success and FALSE otherwise
   return(fResult);
}

//******************************************************************************************
// FUNCTION   Xpm_Pmp_Ifc_SynchronizePrinter
// ABSTRACT   Synchronizes printer by reading pending data and sending synchronization characters
//******************************************************************************************

bool Xpm_Pmp_Ifc_SynchronizePrinter
   (P_XPM_PLUGIN_DATA psData)
{
   bool  fResult                  = false;                      // Return value, FALSE by default
   BYTE  aubEcSynchronize[]       = {0x00,0x00,0x00,0x00,0x00}; // Null bytes, used for synchronization
   DWORD dwWritten, dwRead;                                     // Nr of bytes written/read
   BYTE  ubByte;                                                // Single byte 

   //-- Check parameter
   if (!psData || (psData->dwSructId != P_XPM_PLUGIN_DATA_ID))
      goto FAILURE;

   //-- Read any data that the printer transmits
   dwRead = 0;
   while (psData->pfnReadPort(psData->hPmCallback, &ubByte, sizeof(ubByte), &dwRead, 0) && dwRead);

   //-- Write synchronization bytes to printer
   //.. Note that the timeout for this write is defined in the windows registry key
   //..   "HKLM\Software\Microsoft\WindowsNT\CurrentVersion\Windows\TransmissionRetryTimeout"
   //.. The resolution of this registry key is seconds and the default value is 15 sec.
   dwWritten = 0;
   if (!psData->pfnWritePort(psData->hPmCallback, aubEcSynchronize, sizeof(aubEcSynchronize), &dwWritten, 5000) || (dwWritten != sizeof(aubEcSynchronize)))
      goto FAILURE;  
 
   //-- Wait ... ms 
   Sleep(250);

   //-- Read any data that the printer transmits
   dwRead = 0;
   while (psData->pfnReadPort(psData->hPmCallback, &ubByte, sizeof(ubByte), &dwRead, 0) && dwRead);

   //-- Success
   fResult = true;

   //-- All done
   goto EXIT;

FAILURE:
   ;

EXIT:

   //-- Return TRUE on success and FALSE otherwise
   return(fResult);
}
